//*********************************************************
// Copyright 2014 Intel Corporation All Rights Reserved.
// 
// The source code, information and material ("Material") contained herein is owned by Intel
// Corporation or its suppliers or licensors, and title to such Material remains with Intel 
// Corporation or its suppliers or licensors. The Material contains proprietary information 
// of Intel or its suppliers and licensors. The Material is protected by worldwide copyright 
// laws and treaty provisions. No part of the Material may be used, copied, reproduced, 
// modified, published, uploaded, posted, transmitted, distributed or disclosed in any way 
// without Intel's prior express written permission. No license under any patent, copyright 
// or other intellectual property rights in the Material is granted to or conferred upon 
// you, either expressly, by implication, inducement, estoppel or otherwise. Any license 
// under such intellectual property rights must be express and approved by Intel in writing.
// 
// Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or 
// any other notice embedded in Materials by Intel or Intels suppliers or licensors in any 
// way.
//
// C_OCL.cpp
// Implementation of the C_OCL class
//
//*********************************************************



// External Implementation of C_OCL methods
#include	"C_OCL.h"



////////////////////////PUBLIC METHODS/////////////////////////////



C_OCL::C_OCL( cl_device_type  envTypes)
{

	PlatformNames = (char * *)NULL;
	SupportedPlatformName = NULL;
	SetCurPlatformName( DEFAULT_PLATFORM_NAME );

	CurPlatform = NULL;


	ReadAllPlatforms( );

	AllowedDeviceTypes = envTypes;
	CurDeviceType = CL_DEVICE_TYPE_GPU;		// default only - assumes this is an allowed device type

	CurDevice = NULL;
	DeviceIDs[CPU_INDEX] = NULL;
	DeviceIDs[GPU_INDEX] = NULL;

	CurExtensionString = NULL;

	CurContext = NULL;

	CurQueue = NULL;

	Programs = NULL;
	ProgCount = 0;
	CurMaxProgs = 0;

	Kernels = NULL; 
	KernCount = 0;
	CurMaxKerns = 1;

	return;
}



C_OCL::~C_OCL()
{
	ClearAllPlatforms( );
}



cl_int		C_OCL::Start()
{
	cl_uint		numDevice;
	cl_int		err;

	err = MatchPlatformName( SupportedPlatformName );
	if( err == OCL_ERROR )
	{
		printf( "Platform not found: %s \n", SupportedPlatformName );
		exit(-1);
	}

		// defensive clean up of devices, in case of re-start
	if( DeviceIDs[CPU_INDEX] != NULL )
		clReleaseDevice( DeviceIDs[CPU_INDEX] );

	if( DeviceIDs[GPU_INDEX] != NULL )
		clReleaseDevice( DeviceIDs[GPU_INDEX] );


		// Currently allow only one of each device type per platform, CPU and GPU

	if( AllowedDeviceTypes & CL_DEVICE_TYPE_CPU )
	{
		err = clGetDeviceIDs( CurPlatform, CL_DEVICE_TYPE_CPU, 1, &(DeviceIDs[CPU_INDEX]), &numDevice );
		CL_ERR_CHECK_RETURN(err);
	}

	if( AllowedDeviceTypes & CL_DEVICE_TYPE_GPU )
	{
		err = clGetDeviceIDs( CurPlatform, CL_DEVICE_TYPE_GPU, 1, &(DeviceIDs[GPU_INDEX]), &numDevice );
		CL_ERR_CHECK_RETURN(err);
	}
	

		// set initial device, now that device IDs are set, using currently selected device type
	err = SetCurrentDeviceType( CurDeviceType );


	return	err;
}



	// Support one CPU (host) and one GPU device simultaneous, 
	//	but only one can be selected for running at a time.
	//	This function selects which.  Only CPU_ONLY and GPU_ONLY parameters are allowed.

cl_int		C_OCL::SetCurrentDeviceType( cl_device_type proc )
{

	cl_int	err;

		// device type must be one or other of CPU or GPU
	if( proc == GPU_CPU )
		return	OCL_ERROR;

		// If only one of CPU/GPU is allowed, make sure set proc type matches that
	if( AllowedDeviceTypes != GPU_CPU )
	{
		if( AllowedDeviceTypes != proc )
			return  OCL_ERROR;
	}

	CurDeviceType = proc;

		// NOTE:  if called before "C_OCL::start", the following set CurDevice to NULL
		//	Must call again after initially getting device ids.
	if( CurDeviceType == CL_DEVICE_TYPE_CPU )
		CurDevice = DeviceIDs[CPU_INDEX];
	else if( CurDeviceType == CL_DEVICE_TYPE_GPU )
		CurDevice = DeviceIDs[GPU_INDEX];
	else
		return	OCL_ERROR;


		// NULL current device - return if called before devices are identified (e.g. in arg parsing)
		//	Note: This would happen only if DeviceIDs[] is still uninitialized, i.e. containing NULL
	if( CurDevice == NULL )
		return	OCL_OK;


		// read the device extensions for the new current device
	ReadExtensions( );


		// Create context to work on, associated with the Current Device and its platform
    cl_context_properties context_properties[3] = {CL_CONTEXT_PLATFORM, (cl_context_properties)CurPlatform, NULL };

	if ( CurContext != NULL )
		clReleaseContext( CurContext );

	CurContext = clCreateContext( context_properties, 1, &CurDevice, NULL, NULL, &err );
		CL_ERR_CHECK_RETURN( err );



	if( CurQueue != NULL )
		clReleaseCommandQueue( CurQueue );

		// create a default command queue on current device, context
	CurQueue = clCreateCommandQueue( CurContext, CurDevice, NULL, &err );
		CL_ERR_CHECK_RETURN(err );


	return	CL_SUCCESS;
};
	


	// Check if OpenCL extension named in string "chkstr" is supported on currently selected device

cl_int	C_OCL::CheckExtension( char * chkstr, bool * b_supported )
{
	if( CurExtensionString == NULL || chkstr == NULL )
		return	OCL_ERROR;

	*b_supported	= ((char *)NULL == strstr( CurExtensionString, chkstr ));

	return	CL_SUCCESS;
}



	// First of two functions to retrieve OpenCL device information.
	// This variation expects the user to allocate the correct amount of space for retData to contain returned data
	//		Pass Device Information Constant for devIndex, 
	//		Pass a pointer variable to sufficient space return information on
	//		Pass the size of the data pointed to by retData on retSize - must be large enough or will return error
	// Common use - reading  size_t and other fixed size information

cl_int		C_OCL::GetDeviceInfo( cl_device_info  devInfo, void *  retData,	size_t retSize )
{
	cl_int	err;
	size_t	infoSize;


	err = clGetDeviceInfo( CurDevice, devInfo, 0, NULL,  &infoSize );
		CL_ERR_CHECK_RETURN(err);

	if( retData == NULL || retSize < infoSize )
	{
		return	OCL_ERROR;
	}
	
	err = clGetDeviceInfo( CurDevice, devInfo, infoSize, retData,  NULL );
		CL_ERR_CHECK_RETURN(err);


	return	CL_SUCCESS;
}



	// Second of two functions to retrieve OpenCL device information.
	//	This variation will malloc space to return results 
	//		IMPORTANT NOTE: CALLER MUST free() RETURNED DATA malloc'd BY THIS FUNCTION !
	//		Pass Device Information Constant for devIndex, 
	//		Note that caller will need to cast retData to type returned, per OpenCL spec for clGetDeviceInfo
	//	Common use - reading variable sized information

cl_int		C_OCL::GetDeviceInfo( cl_device_info  devInfo,	void * * retData, size_t *retSize )
{
	cl_int	err;


	if( retData == NULL || retSize == NULL )
		return	OCL_ERROR;

	err = clGetDeviceInfo( CurDevice, devInfo, 0, NULL,  retSize );
		CL_ERR_CHECK_RETURN(err);

	*retData = malloc( *retSize );
	
	err = clGetDeviceInfo( CurDevice, devInfo, *retSize, *retData,  retSize );
		CL_ERR_CHECK_RETURN(err);


	return	CL_SUCCESS;
}



	// Read an UTF-8 coded file.  Using _TCHAR for file names and mode strings and printing errors

int		C_OCL::ReadSourceFile(const char * fileName, char * * retSrc )
{
    FILE *file = NULL;
	int	errCode;
	errCode = fopen_s(&file, fileName, "rb");
    if (errCode != 0)
    {
        printf("ERROR: Failed to open file '%s', error %d\n", fileName, errCode );
        return OCL_ERROR;
    }
	

    if (fseek(file, 0, SEEK_END))
    {
        printf("ERROR: Failed to seek file '%s'\n", fileName);
		fclose(file);
        return OCL_ERROR;
    }

    long size = ftell(file);
    if (size == 0)
    {
        printf("ERROR: Failed to check position on file '%s'\n", fileName);
		fclose(file);
        return OCL_ERROR;
    }

    rewind(file);


    char *src = (char *)malloc(sizeof(char) * size + 1);


    if (!src)
    {
        printf("ERROR: Failed to allocate memory for file '%s'\n", fileName);
		fclose(file);
        return OCL_ERROR;
    }

		//	printf("Reading file '%s' (size %ld bytes)\n", fileName, size);
    size_t res = fread(src, 1, sizeof(char) * size, file);
    if (res != sizeof(char) * size)
    {
        printf("ERROR: Failed to read file '%s'\n", fileName);
		fclose(file);
		free(src);
        return OCL_ERROR;
    }

    src[size] = '\0'; /* NULL terminated */
    fclose(file);

    *retSrc =  src;

	return	CL_SUCCESS;

}



	// Create a program from a file and return it - C_OCL tracks and will release program at end

cl_program	C_OCL::CreateProgramFromFile( char * kernelFileName, int * errRet )	
{
	cl_program	prog;
	int			err;
	cl_int		clErr;

	*errRet = OCL_ERROR;	// default error

	if( Programs == NULL )
	{
		// allocate initial space
		ProgCount = 0;
		CurMaxProgs = 1;

		Programs = (cl_program *) malloc( CurMaxProgs * sizeof(cl_program) );
	}

		// If out of space, create more
	if( ProgCount == CurMaxProgs )
	{
		CurMaxProgs += 8;

		Programs = (cl_program *) realloc( Programs, CurMaxProgs * sizeof(cl_program) );
	}


		//  Validate file, Read source from file to null terminated string
	char *	srcString;
	err = ReadSourceFile( kernelFileName, &srcString );
		if( err != OCL_OK )  {	*errRet = err;  return NULL;	}


		// Create 1 program from null terminated string
    prog = clCreateProgramWithSource(CurContext, 1, (const char **)&srcString, NULL, &clErr);
    if (clErr != CL_SUCCESS )
    {
        printf("ERROR: Failed to create Program with source - error %d\n", clErr );
 
		free(srcString);

		return NULL;
    }
	
		// Build the program
	clErr = clBuildProgram(prog, 1, &CurDevice, "-cl-fast-relaxed-math", NULL, NULL);
    if (clErr != CL_SUCCESS)
    {
        printf("ERROR: Failed to build program - error %d\n", clErr );

		free(srcString);
		clReleaseProgram( prog );
 
		*errRet = err;  
		return NULL;
    }
	
	Programs[ProgCount] = prog;
	ProgCount++;

	free( srcString );

	*errRet = CL_SUCCESS;	// Success!

	return	prog;
}



	// Create a kernel and return handle to it - C_OCL tracks and will release at end

cl_kernel	C_OCL::CreateKernelFromProgram( cl_program  prog, char	* kernelName,  int * errRet  )
{
	cl_kernel	kern;
	cl_int		clErr;

	*errRet = OCL_ERROR;		// default error return

		// Initialize space to store all kernel
	if( Kernels == NULL )
	{
		// allocate initial space
		KernCount = 0;
		CurMaxKerns = 1;		// common test case of one kernel

		Kernels = (cl_kernel *) malloc( CurMaxKerns * sizeof(cl_kernel) );
	}

		// If out of kernel space, create more
	if( KernCount == CurMaxKerns )
	{
		CurMaxKerns += 8;		// if more than one program, assume may be many

		Kernels = (cl_kernel *) realloc( Programs, CurMaxKerns * sizeof(cl_kernel) );
	}

	// Create Kernel here from program and kernel name
	kern = clCreateKernel(prog, kernelName, &clErr);

    if (clErr != CL_SUCCESS)
    {
        printf("ERROR: Failed to create kernel - error %d\n", clErr );
 
		return NULL;
    }

	Kernels[KernCount] = kern;
	KernCount++;

	*errRet = CL_SUCCESS;

	return	kern;
}



	// Clear all created programs - public in case application wishes to do this
	//		But C_OCL will take care of it by default when C_OCL instance is deleted

void	C_OCL::ClearAllPrograms( )
{
	int	index = 0;

	for( index = 0; index < ProgCount; index++ )
	{
		clReleaseProgram( Programs[index] );
	}

	CurMaxProgs = 0;
	ProgCount = 0;

	free( Programs );

	Programs = NULL;
}



	// Clear all created kernels - public in case application wishes to do this
	//		But C_OCL will take care of it by default when C_OCL instance is deleted

void	C_OCL::ClearAllKernels( )
{
	int	index = 0;

	for( index = 0; index < KernCount; index++ )
	{
		clReleaseKernel( Kernels[index] );
	}

	CurMaxKerns = 0;
	KernCount = 0;

	free( Kernels );

	Kernels = NULL;
}




////////////////////////PRIVATE METHODS/////////////////////////////


	// Read all the names of available platforms into class array allocated here

cl_int		C_OCL::ReadAllPlatforms( )
{
	cl_uint		numPlats = 0;
	cl_int		err;
	char  *	*	platNames;

	err = clGetPlatformIDs(0, NULL, &numPlats);
		CL_ERR_CHECK_RETURN(err);

	NumPlatforms = numPlats;

	this->PlatformIDs = (cl_platform_id *) malloc( numPlats * sizeof( cl_platform_id ) );

	err = clGetPlatformIDs(numPlats, this->PlatformIDs, &numPlats);
		CL_ERR_CHECK_RETURN(err);


	PlatformNames = (char **) malloc( numPlats * sizeof( char * ) );
	platNames = PlatformNames;

		// Get list of platform names, save in list of names
	for( cl_uint plat = 0; plat < numPlats; plat++ )
	{
		size_t	size;

		err = clGetPlatformInfo(this->PlatformIDs[plat], CL_PLATFORM_NAME, 0, NULL, &size);
		CL_ERR_CHECK_RETURN(err);

		platNames[plat] = (char *) malloc( size );

		err = clGetPlatformInfo(this->PlatformIDs[plat], CL_PLATFORM_NAME, size, platNames[plat], NULL );
		CL_ERR_CHECK_RETURN(err);
	}

	return	CL_SUCCESS;
}



	// Wipe out all platform-specific information

cl_int		C_OCL::ClearAllPlatforms( )
{
	for( int index = 0; index < NumPlatforms; index++ )
		free( PlatformNames[index]  );

	free( PlatformNames );
	PlatformNames = NULL;

	free( SupportedPlatformName );
	SupportedPlatformName = NULL;

	free(PlatformIDs);
	PlatformIDs = NULL;
	NumPlatforms = 0;

	free( CurExtensionString );
	CurExtensionString  = NULL;

		// release kernels, programs, queue, context, devices
	ClearAllKernels( );
	ClearAllPrograms( );

	clReleaseCommandQueue( CurQueue );
	clReleaseContext( CurContext );
	clReleaseDevice( DeviceIDs[0] );
	clReleaseDevice( DeviceIDs[1] );

	CurQueue = NULL;
	CurContext = NULL;
	CurDevice = DeviceIDs[0] = DeviceIDs[1] = NULL;


	return	CL_SUCCESS;
}


	// Return an index into PlatformIDs[] for the given platform name, or error if name not found

cl_int		C_OCL::MatchPlatformName( char * platName )
{
	int			index;

	for( index = 0 ; index < NumPlatforms; index++ )
	{
		if( 0 == strcmp( PlatformNames[index], platName ) )
		{
			CurPlatform = PlatformIDs[index];
			return	CL_SUCCESS;
		}
	}

	return	OCL_ERROR;
}


		// Read and save the complete extension string for the CurDevice - for use use in CheckExtension() 

cl_int		C_OCL::ReadExtensions( )
{
	cl_int	clErr;
	size_t	sizeExt = 0;

	if( CurExtensionString != NULL )
		free( CurExtensionString );

	clErr = clGetDeviceInfo( CurDevice, CL_DEVICE_EXTENSIONS, 0, NULL, &sizeExt );
		CL_ERR_CHECK_RETURN(clErr);

	CurExtensionString = (char *) malloc( sizeExt );

	clErr = clGetDeviceInfo( CurDevice, CL_DEVICE_EXTENSIONS, sizeExt, CurExtensionString, &sizeExt );
		CL_ERR_CHECK_RETURN(clErr);

	return	CL_SUCCESS;
}


