//*********************************************************
// Copyright 2013-2014 Intel Corporation All Rights Reserved.
// 
// The source code, information and material ("Material") contained herein is owned by Intel
// Corporation or its suppliers or licensors, and title to such Material remains with Intel 
// Corporation or its suppliers or licensors. The Material contains proprietary information 
// of Intel or its suppliers and licensors. The Material is protected by worldwide copyright 
// laws and treaty provisions. No part of the Material may be used, copied, reproduced, 
// modified, published, uploaded, posted, transmitted, distributed or disclosed in any way 
// without Intel's prior express written permission. No license under any patent, copyright 
// or other intellectual property rights in the Material is granted to or conferred upon 
// you, either expressly, by implication, inducement, estoppel or otherwise. Any license 
// under such intellectual property rights must be express and approved by Intel in writing.
// 
// Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or 
// any other notice embedded in Materials by Intel or Intels suppliers or licensors in any 
// way.
//
// ToneMappling.cl
// Implementation of the ToneMappingLine and ToneMapping OpenCL kernels
//		This code is from the Intel(R) SDK for OpenCL Applications
//
//*********************************************************



typedef struct 
{
    float fPowKLow;                             //fPowKLow = pow( 2.0f, kLow)
    float fPowKHigh;                            //fPowKHigh = pow( 2.0f, kHigh)
    float fPow35;                               //fPow35 = pow( 2.0f, 3.5f)
    float fFStops;                              //F stops
    float fFStopsInv;                           //Invesrse fFStops value
    float fPowExposure;                         //fPowExposure = pow( 2.0f, exposure +  2.47393f )
    float fGamma;                               //Gamma correction parameter
    float fPowGamma;                            //Scale factor
    float fDefog;                               //Defog value
} CHDRData;

// "__attribute__((vec_type_hint(float4)))" can be enabled to avoid compiler auto-vectorization

__kernel void /*__attribute__((vec_type_hint(float4)))*/  ToneMappingLine(
	__global float4* inputImage, 
	__global float4* outputImage,
    __global CHDRData* pData,
    int iImageWidth,
    int iImageHeight
    )
{
	// Load tone mapping parameters
    float4 fPowKLow = (float4)pData->fPowKLow;
    float4 fFStops = (float4)pData->fFStops;
    float4 fFStopsInv = (float4)pData->fFStopsInv;
    float4 fPowExposure = (float4)pData->fPowExposure;
    float4 fDefog = (float4)pData->fDefog;
    float4 fGamma = (float4)pData->fGamma;
    float4 fPowGamma = (float4)pData->fPowGamma;


	// and define method constants.
    float4 fOne = 1.0f;
    float4 fZero = 0.0f;
    float4 fSaturate = 255.f;
    
    
    int iCol = get_global_id(0);
    

    for (int iRow = 0; iRow< iImageHeight; iRow++)
    {
        float4 fColor = 0.0f;

        fColor = inputImage[iRow*iImageWidth+iCol];


        //Defog 
        fColor = fColor - fDefog;
        fColor = max(fZero, fColor);


        // Multiply color by pow( 2.0f, exposure +  2.47393f )
        fColor = fColor * fPowExposure;

        int4 iCmpFlag = 0;
        //iCmpFlag = isgreater(fColor, fPowKLow);
        iCmpFlag = fColor > fPowKLow;

        if(any(iCmpFlag))
        {

            //fPowKLow = 2^kLow
            //fFStopsInv = 1/fFStops;
            //fTmpPixel = fPowKLow + log((fTmpPixel-fPowKLow) * fFStops + 1.0f)*fFStopsInv;
            float4 fTmpPixel = fColor - fPowKLow;
            fTmpPixel = fTmpPixel * fFStops;
            fTmpPixel = fTmpPixel + fOne;
            fTmpPixel = native_log( fTmpPixel );
            fTmpPixel = fTmpPixel * fFStopsInv;
            fTmpPixel = fTmpPixel + fPowKLow;

			//color channels update versions
            ///fColor = select(fTmpPixel, fColor, iCmpFlag);
            ///fColor = select(fColor, fTmpPixel, iCmpFlag);
            fColor = fTmpPixel;
        }

        //Gamma correction
        fColor = powr(fColor, fGamma);

        // Scale the values
        fColor = fColor * fSaturate;
        fColor = fColor * fPowGamma;

        //Saturate
        fColor = max(fColor, 0.f);
        fColor = min(fColor, fSaturate);

        //Store result pixel 
        outputImage[iRow*iImageWidth+iCol] = fColor;
    }
}



__kernel void /*__attribute__((vec_type_hint(float4)))*/  ToneMappingPerPixel(
	__global float4* inputImage, 
	__global float4* outputImage,
    __global CHDRData* pData,
    int iImageWidth
    )
{
	// Load tone mapping parameters
    float4 fPowKLow = (float4)pData->fPowKLow;
    float4 fFStops = (float4)pData->fFStops;
    float4 fFStopsInv = (float4)pData->fFStopsInv;
    float4 fPowExposure = (float4)pData->fPowExposure;
    float4 fDefog = (float4)pData->fDefog;
    float4 fGamma = (float4)pData->fGamma;
    float4 fPowGamma = (float4)pData->fPowGamma;


	// and define method constants.
    float4 fOne = 1.0f;
    float4 fZero = 0.0f;
    float4 fSaturate = 255.f;
    
    
    int iCol = get_global_id(0);
    int iRow = get_global_id(1);

    {
        float4 fColor = 0.0f;

        fColor = inputImage[iRow*iImageWidth+iCol];


        //Defog 
        fColor = fColor - fDefog;
        fColor = max(fZero, fColor);


        // Multiply color by pow( 2.0f, exposure +  2.47393f )
        fColor = fColor * fPowExposure;

        int4 iCmpFlag = 0;
        //iCmpFlag = isgreater(fColor, fPowKLow);
        iCmpFlag = fColor > fPowKLow;

        if(any(iCmpFlag))
        {

            //fPowKLow = 2^kLow
            //fFStopsInv = 1/fFStops;
            //fTmpPixel = fPowKLow + log((fTmpPixel-fPowKLow) * fFStops + 1.0f)*fFStopsInv;
            float4 fTmpPixel = fColor - fPowKLow;
            fTmpPixel = fTmpPixel * fFStops;
            fTmpPixel = fTmpPixel + fOne;
            fTmpPixel = native_log( fTmpPixel );
            fTmpPixel = fTmpPixel * fFStopsInv;
            fTmpPixel = fTmpPixel + fPowKLow;

			//color channels update versions
            ///fColor = select(fTmpPixel, fColor, iCmpFlag);
            ///fColor = select(fColor, fTmpPixel, iCmpFlag);
            fColor = fTmpPixel;
        }

        //Gamma correction
        fColor = powr(fColor, fGamma);

        // Scale the values
        fColor = fColor * fSaturate;
        fColor = fColor * fPowGamma;

        //Saturate
        fColor = max(fColor, 0.f);
        fColor = min(fColor, fSaturate);

        //Store result pixel 
        outputImage[iRow*iImageWidth+iCol] = fColor;
    }
}
